/* ++++++++++
	cbserial.c
	A Cardbus 16550A serial port driver (not handling IOctls)
	Written by Martijn Coenen <M.F.A.Coenen@student.tue.nl>
	(C) 2001,2002
	
	Portions of code are (C) Dean Ritchie
	Header file taken from linux sources
	Licensed under the GPL
+++++ */

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ISA.h>
#include <PCI.h>
#include <termios.h>
#include <malloc.h>
#include <cb_enabler.h>
#include "cbserial.h"

#define read8(io_addr)			(*isa->read_io_8)(io_addr)
#define write8(io_addr, val)	(*isa->write_io_8)(io_addr, val)

#define OUTBUFSIZE	8192	
#define INBUFSIZE	8192
#define VENDOR_XIRCOM   0x115d
#define MAX_DEVS 8
#define DEVNAME_LENGTH		64			
#define MAX_CARDS 			 4
#define kDevDir "ports/" 
#define kDevName "kserial32"


int32   api_version     = B_CUR_DRIVER_API_VERSION; 

struct uart_dev_t { 
    char    		*name; 
    uint    		ioport, 
            		irq, 
            		divisor, 
            		nopen; 
    bool   			exists; 
    struct termios 	tty;
    int32			unread_bytes;
    sem_id			read_semaphore;
    sem_id			read_ioctl_semaphore;
    sem_id			write_semaphore;
    char			*inbuf;
 	char			*outbuf;
    int				inrp, inwp;
    int				outrp, outwp;
}; 
 
struct client { 
    uint32            	flags; 
    struct uart_dev_t  	*d; 
}; 

cb_enabler_module_info *cb = NULL;

cb_device_descriptor cb_descrs[] = {
	{ 0x115d, 0x0003, 0xff, 0xff, 0xff }
};

static isa_module_info  *isa; 
static pci_module_info	*pci;

#define MAXPORTS 8

static struct uart_dev_t uart_devices[MAXPORTS];
static int devicecount = 0;

static status_t check_uart (uint32 port);
static status_t init_uart (struct uart_dev_t uart);
int32    serial_interrupt(void *cookie);
status_t match_pci_info(pci_info * item, uint32 *devIDptr);

static status_t
cb_device_added(pci_info *dev, void **cookie)
{
	return match_pci_info(dev, (uint32 *)cookie);
}

static void
cb_device_removed(void *cookie)
{
	uint32 devID = (uint32) cookie;
	int32 previous_state;
}

cb_notify_hooks notify_hooks =
{
	cb_device_added,
	cb_device_removed,
};


/* ----------
	init_hardware - called once the first time the driver is loaded
----- */
status_t
init_hardware (void)
{
	return B_OK;
}

static status_t
check_uart (uint32 port)
{
	uint8 		b, b2, restore;
	uint8		fifo = 0;
	uint8		iirval = 0;
	
  	/* existance test */
 	restore = read8 (port + UART_IER);
  	write8 (port + UART_IER, 0x0);
  	write8 (0x80, 0xff); /* delay */
  	
  	b = read8 (port + UART_IER);
  	write8 (port + UART_IER, 0x0f);
  	write8 (0x80, 0x00); /* delay */
  
  	b2 = read8 (port + UART_IER);
  	write8 (port + UART_IER, restore); /* restore in case not an UART */
  
  	if (b || b2 != 0x0f) {
    	return 0;
	}
  	// Try and enable the FIFO
  	fifo = UART_FCR_ENABLE_FIFO;
  	write8 (port + UART_FCR, fifo);
	
	// Read from the IIR
	iirval = read8(port + UART_IIR);
	if ((iirval >> 6) == 0x03) //FIFO enabled
		return 1;	
  	else
  		return 0;
}


/* ----------
	init_driver - optional function - called every time the driver
	is loaded.
----- */
status_t
init_driver (void)
{
	status_t		status;
	
	if((status = get_module( B_ISA_MODULE_NAME, (module_info **) &isa)) != B_OK) {
		dprintf("Get module failed! %s\n", strerror(status ));
		return status;
	}

	if((status = get_module( B_PCI_MODULE_NAME, (module_info **)&pci )) != B_OK) {
		dprintf("Get module failed! %s\n", strerror(status ));
		return status;
	}
	
	
	if(get_module(CB_ENABLER_MODULE_NAME, (module_info**) &cb) != B_OK) {
		cb = NULL;
	} else {
		cb->register_driver(kDevName, cb_descrs, 1);
		cb->install_notify(kDevName, &notify_hooks);
	}
	
	return B_OK;
}



/* ----------
	uninit_driver - optional function - called every time the driver
	is unloaded
----- */
void
uninit_driver (void)
{
	put_module( B_ISA_MODULE_NAME);
	put_module( B_PCI_MODULE_NAME);
	
	if(cb) {
		cb->uninstall_notify(kDevName, &notify_hooks);
		put_module(CB_ENABLER_MODULE_NAME);
	}

}
	
/* ----------
	serial_open - handle open() calls
----- */

static status_t
serial_open (const char *name, uint32 flags, void** cookie)
{
	struct client	*c;	
	int				i = 0;
	bool			found = false;
	uint16			baud_value;
	char			devname[64];
			
	while (i < devicecount && !found) {
		if (strcmp( name, uart_devices[i].name) == 0)
			found = true;
		else 
			i++;
	}
	if (!found)
		return B_ERROR;
	
	c = malloc( sizeof *c); 
    
    if (c == 0) 
        return (ENOMEM); 
    
    *cookie = c; 
    
    c->flags = flags; 
    
    c->d = &(uart_devices[i]); 		
    c->d->unread_bytes = 0;
  	
  	c->d->inrp = 0;
  	c->d->inwp = 0;
  	c->d->outrp = 0;
  	c->d->outwp = 0;
  	
  	c->d->read_semaphore = create_sem(0, "BeSerial");
  	c->d->read_ioctl_semaphore = create_sem(0, "BeSerial");
  	c->d->write_semaphore = create_sem(1, "BeSerial");	
	
	c->d->inbuf = malloc(INBUFSIZE);
	c->d->outbuf = malloc(OUTBUFSIZE);
	
	install_io_interrupt_handler(c->d->irq, serial_interrupt, (c->d), 0 ); 
	
	write8(c->d->ioport + UART_FCR, UART_FCR_ENABLE_FIFO);
	write8(c->d->ioport + UART_FCR, UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT);
	
	write8(c->d->ioport + UART_LCR, UART_LCR_WLEN8);	/* reset DLAB */
	
	// Clear interrupts

	read8(c->d->ioport + UART_LSR);
	read8(c->d->ioport + UART_RX);
	read8(c->d->ioport + UART_IIR);
	read8(c->d->ioport + UART_MSR);

	// Set DLAB access
	write8(c->d->ioport + UART_LCR, UART_LCR_DLAB);
	
	// Set baudrate
	baud_value = 115200 / B115200;
  	write8(c->d->ioport + UART_DLL, baud_value & 0xff);
  	write8(c->d->ioport + UART_DLM, (baud_value >> 8) & 0xff);
  
  	/* set parity, stop bits, data length */
  	write8(c->d->ioport + UART_LCR, UART_LCR_WLEN8);

	// Enable fifo
	
	write8(c->d->ioport + UART_FCR, UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT | UART_FCR_TRIGGER_8);
	
	// Enable interrupts
	
	write8(c->d->ioport + UART_IER, UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI);
	write8(c->d->ioport + UART_MCR, UART_MCR_OUT2 | UART_MCR_DTR | UART_MCR_RTS);	
		
	return B_OK;
}


/* ----------
	serial_read - handle read() calls
----- */

static status_t
serial_read (void* cookie, off_t position, void *buf, size_t* num_bytes)
{
	int32		count = 0;
	struct client *uart_client = (struct client *)cookie;
	struct uart_dev_t *dev = uart_client->d ;
	char *msgbuf = (char *) buf;
	int			status;
		
	if (uart_client->flags & O_NONBLOCK) {
		status = acquire_sem_etc(dev->read_semaphore,1, B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT, 0);
		if (status == B_WOULD_BLOCK) {
			return(B_WOULD_BLOCK);
		}
	}
	else {
		acquire_sem_etc(dev->read_semaphore,1, B_CAN_INTERRUPT, 0);
	}

	if (dev->inwp < dev->inrp) {
		/* The pointer flipped over the boundary */
		while (dev->inrp != 0) {
			msgbuf[count] = dev->inbuf[dev->inrp];
			dev->inrp = ((dev->inrp) + 1) % 8192;
			count++;
			dev->unread_bytes--;
		}
	} 
	while (dev->inrp != dev->inwp) {
		msgbuf[count] = dev->inbuf[dev->inrp];
		dev->inrp = ((dev->inrp) + 1) % 8192;
		count++;
		dev->unread_bytes--;
	}

	*num_bytes = count;	
	
	return B_OK;
}


/* ----------
	serial_write - handle write() calls
----- */

static status_t
serial_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	char *charbuf = (char *)buffer;
	int32		written = 0;	
	struct client *uart_client = (struct client *)cookie;
	struct uart_dev_t *dev = uart_client->d ;
	
  	acquire_sem_etc(dev->write_semaphore, 1, B_CAN_INTERRUPT, 0);
  	
  	while (written < *num_bytes) {
  		dev->outbuf[dev->outwp] = charbuf[written];
		dev->outwp = (dev->outwp + 1) % 8192;	
		written++;
	}
	
	write8(dev->ioport + UART_IER, UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI | UART_IER_THRI);	

	release_sem(dev->write_semaphore);
	
	*num_bytes = written;
	
	return B_OK;
}


/* ----------
	serial_control - handle ioctl calls
----- */

static status_t
serial_control (void* cookie, uint32 op, void* data, size_t len)
{
	// *** No IOCTL's are used yet, they all get called upon the pty
	// If you are interested in a driver directly handling these ioctls (I am developing it)
	// Mail me.
	
	struct client *uart_client = (struct client *)cookie;
	struct uart_dev_t *dev = uart_client->d ;
		  
	switch ( op ) {
	case B_SET_NONBLOCKING_IO: 			// Set to non-blocking i/o
		dprintf("Non blocking ioctl called\n");
		atomic_or(&(uart_client->flags), (uint32) O_NONBLOCK);
		break;

	case B_SET_BLOCKING_IO:				// Set to blocking i/o
		dprintf("Blocking ioctl called\n");
		atomic_and(&(uart_client->flags), (uint32) ~O_NONBLOCK);
		break;
		
	case TCGETBITS:
		{
		int bits;
		
		//DEBUG_ONLY( dprintf(LOGID "--- control_hook %s: TCGETBITS\n", dev->name) );
		dprintf("TCGETBITS ioctl called\n");
		bits = 0;
		// if ( dev->control_signal_state & CDC_CONTROL_SIGNAL_STATE_DTR) bits |= 16;
		// if ( dev->control_signal_state & CDC_CONTROL_SIGNAL_STATE_RTS) bits |= TCGB_CTS;
		// if ( dev->serial_state & CDC_UART_STATE_DSR) bits |= TCGB_DSR;
		// if ( dev->serial_state & CDC_UART_STATE_RING) bits |= TCGB_RI;
		// if ( dev->serial_state & CDC_UART_STATE_DCD) bits |= TCGB_DCD;
		
		*((int *) data) = bits;
		break;
		};

	case TCGETA:
		{
		struct termios *	t;

		dprintf("TCGETA ioctl called\n");	
		t = (struct termios *) data;
		
		memcpy(t, &dev->tty, sizeof(dev->tty));
		return(B_OK);
		
		break;
		};

	case TCSETA:
		{
		struct termios *	t;

		dprintf("TCSETA ioctl called\n");						
		t = (struct termios *) data;
				
		memcpy(&dev->tty, t, sizeof(dev->tty));
		return(B_OK);

		break;
		};

	case TCWAITEVENT:
		{ 
		bigtime_t	timeout;
		
		dprintf("TCWAITEVENT called with VMIN = %d and VTIME = %d\n", dev->tty.c_cc[VMIN], dev->tty.c_cc[VTIME]);
			
		if( dev->tty.c_cc[VTIME] == 0 && dev->tty.c_cc[VMIN] != 0 ) {
			timeout = B_INFINITE_TIMEOUT;
			dprintf("TCWAITEVENT with infinite timeout\n");
		}
		else {
			timeout = 100000 * dev->tty.c_cc[VTIME];
		}
		
		while ( (dev->unread_bytes < dev->tty.c_cc[VMIN]) || (dev->tty.c_cc[VMIN] == 0) )
		{
			int status;
			status = acquire_sem_etc(dev->read_ioctl_semaphore, 1, B_CAN_INTERRUPT | B_RELATIVE_TIMEOUT, timeout);
				
			if ( status == B_WOULD_BLOCK )
			{
				dprintf("Exited with would block\n");
				break;
			};
									
			if ( dev->tty.c_cc[VMIN] == 0 )
				// We only try once, no minimum input is required 
				{
				break;
				};
				
			if ( dev->tty.c_cc[VTIME] != 0 && status == B_TIMED_OUT )
				{
				dprintf("Timed out\n");
				break; // Timeout elapses before enough (VMIN] input come, so exit now
				};

		};
		if ( dev->unread_bytes > 0 )
		{
			// We have data
			if ( data )
				*((int *) data) = EV_CARRIER;
		}
		else {
			*((int *) data) = 0;
		}
		
		return B_OK;
		};

	case TCSETDTR:
		{
		bool	on;
		uint16	control_signal_state;
		
		on = *((bool *) data);
		/*
		control_signal_state = dev->control_signal_state;
		if ( on )
			control_signal_state |= CDC_CONTROL_SIGNAL_STATE_DTR;
		else
			control_signal_state &= ~ CDC_CONTROL_SIGNAL_STATE_DTR;
		*/
		
		//return device_set_control_signal_state(dev, control_signal_state);
		dprintf("TCSETDTR ioctl called\n");
		return B_OK;
		};

	case TCSETRTS:
		{
		bool	on;
		uint16	control_signal_state;
		
		on = *((bool *) data);
		/*
		control_signal_state = dev->control_signal_state;
		if ( on )
			control_signal_state |= CDC_CONTROL_SIGNAL_STATE_RTS;
		else
			control_signal_state &= ~ CDC_CONTROL_SIGNAL_STATE_RTS;
		*/
		
		//return device_set_control_signal_state(dev, control_signal_state);
		dprintf("TCSETRTS ioctl called\n");
		return B_OK;
		};

	case TCFLSH:
		{
		int		output;
		
		output = *((int *) data);
		dprintf("Flush ioctl called\n");
		
		//if ( ! output )
			// flush unread bytes
		//	dev->unread_bytes = 0;
		//};
		break;
		//return B_OK;
		}
	default:
		{
		dprintf("Invalled opcode called\n");
		return EINVAL;
		};
	}
	dprintf("Left ioctl\n");
	return B_OK;
	
}


/* ----------
	serial_close - handle close() calls
----- */

static status_t
serial_close (void* cookie)
{	
	int		status;
	
	struct client *uart_client = (struct client *)cookie;
	struct uart_dev_t *dev = uart_client->d ;
	
	remove_io_interrupt_handler(dev->irq, serial_interrupt, dev);
	return B_OK;
}


/* -----
	serial_free - called after the last device is closed, and after
	all i/o is complete.
----- */
static status_t
serial_free (void* cookie)
{
	//free(cookie);
	return B_OK;
}

int32    serial_interrupt(void *cookie) {
	int32 handled = B_HANDLED_INTERRUPT;
	struct uart_dev_t *dev = (struct uart_dev_t *)cookie;
	int status;
	
	if (read8(dev->ioport + UART_IIR) & UART_IIR_NO_INT)
		return B_UNHANDLED_INTERRUPT;
	
	do {
		int		writecount = 16;
		char	testchar;
		int		msr;
				
		status = read8(dev->ioport + UART_LSR);
		
		if (status & UART_LSR_DR) {
			do {
				dev->inbuf[dev->inwp] =  read8(dev->ioport + UART_RX);
     			dev->inwp = ((dev->inwp + 1) % 8192);
				status = read8(dev->ioport + UART_LSR);
				dev->unread_bytes++;
			} while (status & UART_LSR_DR);
		}
		
		msr = read8(dev->ioport + UART_MSR);
				
		if (status & UART_LSR_THRE) {
	
			if (dev->outwp < dev->outrp) {
				// Boundary flip
				while (dev->outrp != 0 && writecount > 0) {
	    			write8(dev->ioport + UART_TX, dev->outbuf[dev->outrp]); 	
					dev->outrp = (dev->outrp + 1) % 8192;
					writecount--;
				}
			}
			while (dev->outrp < dev->outwp && writecount > 0) {
	    		write8(dev->ioport + UART_TX, dev->outbuf[dev->outrp]);   	    
   				dev->outrp = (dev->outrp + 1) % 8192;
				writecount--;
   			}
		}
		
	} while ((!(read8(dev->ioport + UART_IIR) & UART_IIR_NO_INT)));
	
	if (dev->outwp == dev->outrp)
		write8(dev->ioport + UART_IER, UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI);
	else
		write8(dev->ioport + UART_IER, UART_IER_MSI | UART_IER_RLSI | UART_IER_RDI | UART_IER_THRI);	
	if (dev->unread_bytes) {
		release_sem_etc(dev->read_semaphore, 1, B_DO_NOT_RESCHEDULE);
		
		/* ! */
		release_sem_etc(dev->read_ioctl_semaphore, 1, B_DO_NOT_RESCHEDULE);
		
		return B_INVOKE_SCHEDULER;
	}
	return handled;
}

/* -----
	function pointers for the device hooks entry points
----- */

device_hooks my_device_hooks = {
	serial_open, 			/* -> open entry point */
	serial_close, 			/* -> close entry point */
	serial_free,			/* -> free cookie */
	serial_control, 		/* -> control entry point */
	serial_read,			/* -> read entry point */
	serial_write			/* -> write entry point */
};


/* -----
	null-terminated array of device names supported by this driver
----- */

static const char *my_device_name[] = {
	"ports/kserial",
	NULL
};

static const char *none_found[] = {
	NULL
};

/* ----------
	publish_devices - return a null-terminated array of devices
	supported by this driver.
----- */

const char**
publish_devices()
{
	if (devicecount)
		return my_device_name;
	else
		return none_found;
}

/* ----------
	find_device - return ptr to device hooks structure for a
	given device name
----- */

device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

status_t match_pci_info(pci_info * item, uint32 *devIDptr) {

	pci_info *new_item;
	uint16 free_slot_index;
	int32 previous_state;
	int32	ioport, irq, bus, pcidevice;
	bus = item->bus;
	pcidevice = item->device;
		
	// Try function 1
	ioport = read_pci_config(bus, pcidevice, 1, 0x10, 4) - 1;
	dprintf("Read modem port 0x%04x\n",ioport); 
	irq = read_pci_config(bus, pcidevice, 1, 0x3C,4) & 0xff;
	dprintf("Read modem irq as %d\n", irq);
				
	if (check_uart(ioport)) {
		struct uart_dev_t uart;
	
		uart.ioport = ioport;
		uart.irq = irq;
		uart.name = "ports/kserial";
		uart_devices[devicecount] = uart;
		devicecount++;
	}

	return B_OK;
		
}
